
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%            A code for calibration of model parameters                   % 
%                           developed by                                  %
%                         Peyman Babakhani                                % 
%                    University of Liverpool, UK                          %
%               National Tsing Hua University, Taiwan                     %
%                    p.babakhani@liverpool.ac.uk                          %
%                        peyman.bi@yahoo.com                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                         %
%   This code is based on a heuristic optimization algorithm. This        %
%   is aimed at dealing with potential stability problems of numerical    %
%   models, and also to utilize the parallel computational capability     %
%   more efficiently in the calibration process.                          %
%                                                                         %
%   The "slave" model needs to read the observation data, solves the      %
%   aggregation and sedimentation model, compare observations with        %
%   modelled data and calculate the objective function which here is      %
%   Nash-Sutcliff R^2.                                                    % 
%                                                                         %
%   This algorithm consists of two major loops. In the first loop, the    % 
%   Nash-Sutcliff R^2, as an objective function, is elevated iteratively  % 
%   until the difference between R^2 in two successive iterations reduces % 
%   to below the first tolerance defined by the user. In the second loop, %
%   a similar procedure is pursued with more refined parameter increments % 
%   until the difference between R^2 in two successive iterations reaches % 
%   reaches below the second tolerance defined by the user.               %
%                                                                         %   
%   The code fits the mean hydrodynamic size and normalized derived count % 
%   rate of DLS (DCR), as representative of dimensionless mass            %
%   concentration produced by the model to those of experimental data.    %
%   The weight of contribution of each of these datasets can be modified. %
%   This optimization code, acting as a master model, is designed        % 
%   to decrease the time step of the slave model, if this goes unstable. %
%   Additionally, the scheme considers if after three increments          %
%   in the time step size, the model is still unstable, the optimization  % 
%   algorithm skips that multiplication of the parameter value.           % 
%   This code can only optimize one experimental dataset at a time.       %
%   Further information on this code is available in the following        % 
%   reference:                                                            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!%
%   IMPORTANT NOTE: The user can use and modify this code for any         % 
%   non-commercial purposes with proper reference to the following paper: %
%   Peyman Babakhani, Jonathan Bridge, Tanapon Phenrat, Ruey-an Doong,    %   
%   and Karl R Whittle, 2018 "Aggregation and sedimentation of shattered  % 
%   graphene oxide nanoparticles (SGO) in dynamic environments: a         % 
%   solid-body rotational approach"                                       %
%   Environmental, Science: Nano                                          %
%!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear all;close all;clc   % To clear everything
tic;                      % To start recording the CPU time
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Inputs:
%vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
FileName='Cyl_GO_Case10';
 
%The name of the MATLAB function which contains the slave model
str = 'Fix_Piv_Code_Slave_Function_for_Fitting';
 
%Initial Values of Parameters:
alfinit=0.0072; %Attachment Efficiency 
Dfinit=2.626;   %Fractal Dimension 
%Factors for increasing the initial time steps when the slave model becomes unstable:
% The runs start with the first element and if an unstable run occurs,
% then the second element will be tried and so forth. 
deltfacinMat=[120 1000 2000];
 
%Paramter bounds:
%The first element corresponds to alpha and the second corresponds to Df
ub=[50 2.99];  %Upper bound
lb=[1e-10 1.3]; %Lower bound
 
%Mulitiplying factors in parameters:
SensFacAlf=[0.7 0.8 0.9 1 1.1 1.2 1.3];   %Alpha in the first loop
SensFac_RefinedAlf=[0.95 0.975 0.99 1 1.01 1.025 1.05]; %Df in the first loop
SensFacDf=[0.85 0.90 0.95 1 1.05 1.1 1.15];   %Alpha in the second loop
SensFac_RefinedDf=[0.97 0.98 0.99 1 1.01 1.02 1.03];   %Df in the second loop
 
% Weight factors for each objective function of mean size data and
% concentration data (0 < weight_factor < 1):
% For example, R2Dhwieght=1 and R2Concwieght=0 means that only the data of
% mean size is considered in the optimization process.
R2Dhwieght=1;  %Weight factor for mean size data
R2Concwieght=0.0; %Weight factor for normalized concentration data
 
% Enter the tolerance in which mass balance error of the slave model is
% accceptable(%):
MassErrorTolerance=0.00001;  
 
% Enter the tolerance for R^2 between two successive iterations(%) in Loop 1:
IterToleran1=0.1;
% Enter the tolerance for R^2 between two successive iterations(%) in Loop 2:
IterToleran2=0.01;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %Conversation of excel files:  
%This part converts input excel files for being read by the Slave model in
% a .mat format
 
%First Read the input files (experimental data) which will be used for
%optimization:
Input1=xlsread('Outputs_Polydis_Compelet_Cyl_UV_Vis_300min_075mM_IS.xls', 'Sheet1');
[num,txt,raw] = xlsread('Outputs_Polydis_Compelet_Cyl_UV_Vis_300min_075mM_IS.xls', 'Sheet1'); 
%Reading the count rate data:
Input2=xlsread('Outputs_Polydis_Compelet_Cyl_UV_Vis_300min_075mM_IS.xls', 'Sheet2');
%Reading the size distribution data:
x_initial_size_dist=xlsread('XData_Size_Distrib.xls');
obs_size_dist=xlsread('Outputs_Polydis_Compelet_Cyl_UV_Vis_300min_075mM_IS.xls', 'Sheet3');
 
%Then save the .mat files:
   save('Input1.mat','Input1')
   save('Input2.mat','Input2')
   save('x_initial_size_dist.mat','x_initial_size_dist')
   save('obs_size_dist.mat','obs_size_dist')
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Start of optimization (fitting):
 
fh = str2func(str);
alfcal=alfinit;
Dfcal=Dfinit;
deltfacin=deltfacinMat(1);
 
%Initial run of the slave model:
[Rsqout,RsqConcout,finalMassBalanceErrorout]=fh(alfcal,Dfcal,deltfacin);
Rsqout
RsqConcout
% Exerting the weights for R^2:
if isnan(Rsqout)
    Rsqselc1=-100;  %Initial assumption
else
Rsqselc1=(R2Dhwieght*Rsqout+R2Concwieght*RsqConcout)/(R2Dhwieght+R2Concwieght);
end 
 
Rsqselc2=-50;   %Initial assumption
RsqAlfMat=zeros(7,1);
DfinMat=zeros(7,1);
ca=0;
sizeIncrement=size(deltfacinMat);
 
%**************************************************************************
%Start of the first loop:
%**************************************************************************
while abs(Rsqselc2-Rsqselc1)>IterToleran1
 
Rsqout=0;
 
% sensitivities for alfa:
deltfacin=deltfacinMat(1);
alfinMat=alfcal*SensFacAlf;
alfinMat(alfinMat>ub(1))=[];
alfinMat(alfinMat<lb(1))=[];
elem=size(alfinMat);
elemen=elem(2);
 
parfor j=1:elemen   %Parallel loop
    
deltfacin=deltfacinMat(1);
 
Dfin=Dfcal;
alfin=alfinMat(j);
 
[Rsqout,RsqConcout,finalMassBalanceErrorout]=fh(alfin,Dfin,deltfacin);
 
for jj=1:sizeIncrement(2)
if (isnan(finalMassBalanceErrorout) || isnan(Rsqout) || abs(finalMassBalanceErrorout)>MassErrorTolerance)
    deltfacin=deltfacinMat(jj);
[Rsqout,RsqConcout,finalMassBalanceErrorout]=fh(alfin,Dfin,deltfacin);
end
end
 
RsqAlfMat(j)=(R2Dhwieght*Rsqout+R2Concwieght*RsqConcout)/(R2Dhwieght+R2Concwieght);
RsqConcAlfMat(j)=RsqConcout;
finalMassBalanceErrorAlfMat(j)=finalMassBalanceErrorout;
 
end
 
% Removing NaN from the matrices 
RsqAlfMat(isnan(RsqAlfMat))=-100;
maxRsqAlf=max(RsqAlfMat);
indmaxRsqAlf=find(RsqAlfMat==maxRsqAlf);
 
% sensitivities for Df:
Rsqout=0;
deltfacin=deltfacinMat(1);
DfinMat=Dfcal.*SensFacDf;
DfinMat(DfinMat>ub(2))=[];
DfinMat(DfinMat<lb(2))=[];
elemDf=size(DfinMat);
elemenDf=elemDf(2);
 
parfor jd=1:elemenDf   %Parallel loop
    
deltfacin=deltfacinMat(1);
 
Dfin=DfinMat(jd);
alfin=alfcal;
 
[Rsqout,RsqConcout,finalMassBalanceErrorout]=fh(alfin,Dfin,deltfacin);
 
% This loop alters the slave model time step if it goes unstable:
for jj=1:sizeIncrement(2) 
    
if (isnan(finalMassBalanceErrorout) || isnan(Rsqout) || abs(finalMassBalanceErrorout)>MassErrorTolerance)
 
    deltfacin=deltfacinMat(jj);
[Rsqout,RsqConcout,finalMassBalanceErrorout]=fh(alfin,Dfin,deltfacin)
 
end
end
 
RsqDfMat(jd)=(R2Dhwieght*Rsqout+R2Concwieght*RsqConcout)/(R2Dhwieght+R2Concwieght);
RsqConcDfMat(jd)=RsqConcout;
finalMassBalanceErrorDfMat(jd)=finalMassBalanceErrorout;
 
end
 
% Removing NaN from the matrix 
RsqDfMat(isnan(RsqDfMat))=-100;
maxRsqDf=max(RsqDfMat);
indmaxRsqDf=find(RsqDfMat==maxRsqDf);
 
if maxRsqAlf>=maxRsqDf
alfcal=alfinMat(indmaxRsqAlf)
maxRsqAlf
Rsqselc1=Rsqselc2;
Rsqselc2=maxRsqAlf;
 
else
Dfcal=DfinMat(indmaxRsqDf)
maxRsqDf
Rsqselc1=Rsqselc2;
Rsqselc2=maxRsqDf;
 
end
 
ca=ca+1;
if ca>1000
    break
end
IterationNo=ca
    
end
 
%**************************************************************************
% Continue of optimization with refinement of parameter factors (second loop):
%**************************************************************************
Rsqselc1=-100;
Rsqselc2=-50;
RsqAlfMat=zeros(7,1);
DfinMat=zeros(7,1);
ca2=0;
 
while abs(Rsqselc2-Rsqselc1)>IterToleran2
 
Rsqout=0;
 
% Sensitivities for alfa:
deltfacin=deltfacinMat(1);
 
alfinMat=alfcal.*SensFac_RefinedAlf;
alfinMat(alfinMat>ub(1))=[];
alfinMat(alfinMat<lb(1))=[];
elem=size(alfinMat);
elemen=elem(2);
 
parfor j=1:elemen   %Parallel loop
    
deltfacin=deltfacinMat(1);
 
Dfin=Dfcal;
alfin=alfinMat(j);
 
[Rsqout,RsqConcout,finalMassBalanceErrorout]=fh(alfin,Dfin,deltfacin);
 
for jj=1:sizeIncrement(2)
 
if (isnan(finalMassBalanceErrorout) || isnan(Rsqout) || abs(finalMassBalanceErrorout)>MassErrorTolerance)
    deltfacin=deltfacinMat(jj)
[Rsqout,RsqConcout,finalMassBalanceErrorout]=fh(alfin,Dfin,deltfacin);
end
end
 
RsqAlfMat(j)=(R2Dhwieght*Rsqout+R2Concwieght*RsqConcout)/(R2Dhwieght+R2Concwieght);
RsqConcAlfMat(j)=RsqConcout;
finalMassBalanceErrorAlfMat(j)=finalMassBalanceErrorout;
 
end
 
%Removing NaN from the matrix 
RsqAlfMat(isnan(RsqAlfMat))=-100;
maxRsqAlf=max(RsqAlfMat);
indmaxRsqAlf=find(RsqAlfMat==maxRsqAlf);
 
% Sensitivities for Df:
Rsqout=0;
deltfacin=deltfacinMat(1);
DfinMat=Dfcal.*SensFac_RefinedDf;
DfinMat(DfinMat>ub(2))=[];
DfinMat(DfinMat<lb(2))=[];
elemDf=size(DfinMat);
elemenDf=elemDf(2);
 
parfor jd=1:elemenDf   %Parallel loop
    
deltfacin=deltfacinMat(1);
 
Dfin=DfinMat(jd);
alfin=alfcal;
 
[Rsqout,RsqConcout,finalMassBalanceErrorout]=fh(alfin,Dfin,deltfacin);
 
 
for jj=1:sizeIncrement(2)
    
if (isnan(finalMassBalanceErrorout) || isnan(Rsqout) || abs(finalMassBalanceErrorout)>MassErrorTolerance)
    deltfacin=deltfacinMat(jj)
[Rsqout,RsqConcout,finalMassBalanceErrorout]=fh(alfin,Dfin,deltfacin);
 
end
end
 
RsqDfMat(jd)=(R2Dhwieght*Rsqout+R2Concwieght*RsqConcout)/(R2Dhwieght+R2Concwieght);
RsqConcDfMat(jd)=RsqConcout;
finalMassBalanceErrorDfMat(jd)=finalMassBalanceErrorout;
 
end
 
%Removing NaN from the matrix 
RsqDfMat(isnan(RsqDfMat))=-100;
maxRsqDf=max(RsqDfMat);
indmaxRsqDf=find(RsqDfMat==maxRsqDf);
 
if maxRsqAlf>=maxRsqDf
alfcal=alfinMat(indmaxRsqAlf)
maxRsqAlf
Rsqselc1=Rsqselc2;
Rsqselc2=maxRsqAlf;
 
else
Dfcal=DfinMat(indmaxRsqDf)
maxRsqDf
Rsqselc1=Rsqselc2;
Rsqselc2=maxRsqDf;
 
end
 
ca2=ca2+1;
if ca2>1000
    break
end
IterationNoRefined=ca2
 
end
 
 
%Final Run of the Model:
[Rsqout,RsqConcout,finalMassBalanceErrorout]=fh(alfcal,Dfcal,deltfacin);
Rsqout
RsqConcout
 
save(['Calib_Outcomes_' num2str(FileName)]); %Save the final results
 
FinalResults=[alfcal Dfcal Rsqout RsqConcout]
 
TableItems={'Alpha/Hamaker' 'Df' 'maxRsqAlf' 'maxRsqDf'};
 
dlmwrite(['Calib_Result_Table_' num2str(FileName)],transpose(FinalResults),'delimiter',...
        '\t','precision',16,'roffset',1);
 


